/*-
 * Copyright (c) 2013 Devin Teske <dteske@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "fdpv.h"
#include "dprompt.h"
#include "string_m.h"

/* Mecahnics */
#define INCREMENT		1	/* How much to increment % each time */
#define XDIALOG_INCREMENT	15	/* different for slower Xdialog(1) */
#define SLEEP_USEC		0	/* ms to sleep between increments */
#define XDIALOG_SLEEP_USEC	0	/* different for slower Xdialog(1) */
static u_int8_t increment = INCREMENT;
static useconds_t sleep_usec = SLEEP_USEC;

/* Data to process */
int nfiles = 0;
static int nthfile = 0;
static file_node *file_list = (file_node *)(0);

/* Extra display information */
char *pprompt = (char *)(0);
char *aprompt = (char *)(0);

/* dialog(1) characteristics */
char dialog[PATH_MAX] = DIALOG;
u_int8_t use_dialog = 1, use_xdialog = 0, use_color = 1;
	/* merely defaults -- affected by command-line arguments/environment */

/* Function prototypes */
void usage(void);

int
main(int argc, char *argv[])
{
	int ch;
	int overall = 0;
	file_node *curfile;

	/*
	 * Process command-line options
	 */
	while ((ch = getopt(argc, argv, "L:P:a:p:s:X")) != -1)
	{
		switch(ch) {
		case 'L': /* custom label size */
			label_size =
				(int)strtol(optarg, (char **)NULL, 10);
			if (!label_size && errno == EINVAL)
			{
				fprintf(stderr, "%s: `-L' argument must be "
				        "numeric\n", argv[0]);
				exit(EXIT_FAILURE);
			}
			else if (label_size < 1)
			{
				fprintf(stderr, "%s: `-L' argument must be "
				        "greater than zero\n", argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 'P': /* custom mini progress bar size */
			pbar_size =
				(int)strtol(optarg, (char **)NULL, 10);
			if (!pbar_size && errno == EINVAL)
			{
				fprintf(stderr, "%s: `-P' argument "
				        "must be numeric\n", argv[0]);
				exit(EXIT_FAILURE);
			}
			else if (pbar_size < 4)
			{
				fprintf(stderr, "%s: `-P' argument must be "
				        "greater than four\n", argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 'a': /* additional message text to append */
			if (!aprompt) aprompt =
				(char *)malloc(APROMPT_MAX);
			snprintf(aprompt, APROMPT_MAX, "%s", optarg);
			break;
		case 'p': /* additional message text to use as prefix */
			if (!pprompt) pprompt =
				(char *)malloc(PPROMPT_MAX);
			snprintf(pprompt, PPROMPT_MAX, "%s", optarg);
			break;
		case 's': /* size argument */
			break;
		case 'X': /* X11 support through x11/xdialog */
			snprintf(dialog, PATH_MAX, XDIALOG);
			use_xdialog = 1;
			use_dialog = 0;
			use_color = 0;
			break;
		case '?': /* unknown argument (based on optstring) */
		default: /* unhandled argument (based on switch) */
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	/* Xdialog(1) requires newlines (a) escaped and (b) in triplicate */
	if (use_xdialog)
	{
		int len;

		/* Replace `\n' with `\n\\n\n' in pprompt */
		len = strlen(pprompt);
		len += strcount(pprompt, "\\n") * 2;
		if (len > PPROMPT_MAX)
		{
			fprintf(stderr, "%s:%d: Oops, pprompt buffer overflow"
			        " (%i > %i)\n", __FILE__, __LINE__,
			        len, PPROMPT_MAX);
			exit(EXIT_FAILURE);
		}
		if (replaceall(pprompt, "\\n", "\n\\n\n") < 0)
		{
			fprintf(stderr, "%s:%d: replaceall: %s (%i)\n",
			        __FILE__, __LINE__, strerror(errno), errno);
			exit(EXIT_FAILURE);
		}

		/* Replace `\n' with `\n\\n\n' in aprompt */
		len = strlen(aprompt);
		len += strcount(aprompt, "\\n") * 2;
		if (len > APROMPT_MAX)
		{
			fprintf(stderr, "%s:%d: Oops, aprompt buffer overflow"
			        " (%i > %i)\n", __FILE__, __LINE__,
			        len, APROMPT_MAX);
			exit(EXIT_FAILURE);
		}
		if (replaceall(aprompt, "\\n", "\n\\n\n") < 0)
		{
			fprintf(stderr, "%s:%d: replaceall: %s (%i)\n",
			        __FILE__, __LINE__, strerror(errno), errno);
			exit(EXIT_FAILURE);
		}
	}

	/* Xdialog(1) updates a lot slower than dialog(1) */
	if (use_xdialog) increment = XDIALOG_INCREMENT;

	/* Process remaining arguments as list of file-names */
	{
		int n = 0;
		size_t size = sizeof(file_node);

		for (curfile = file_list; n < argc; n++)
		{
			nfiles++;
			if (!curfile) {
				curfile = (file_node *)malloc(size);
				memset((void *)(curfile), '\0', size);
				file_list = curfile;
			}
			else
			{
				curfile->next = (file_node *)malloc(size);
				memset((void *)(curfile->next), '\0', size);
				curfile = curfile->next;
			}
			curfile->name = argv[n];
		}
	}

	dprompt_init();

	/*
	 * Process the file streams [serially] (one for each argument passed)
	 */
	for (curfile = file_list; curfile; curfile = curfile->next)
	{
		int pct = 0 - increment;

		nthfile++;
		while ((pct += increment) <= 100)
		{
			if (sleep_usec) usleep(sleep_usec);

			/* Calculate total overall progress */
			overall = ( 100 * nthfile - 100 + pct) / nfiles;

			/* Round up baed on one-tenth of a percent */
			if (((100*nthfile - 100 + pct)*10 / nfiles % 100) > 50)
				overall++;

			dprompt_recreate(file_list, curfile, pct);
			dprompt_fprint(stdout, pprompt, aprompt, overall);
			fflush(stdout);
		}
	}

	exit(EXIT_SUCCESS);
}

void
usage(void)
{
	exit(EXIT_FAILURE);
}
