/*-
 * Copyright (c) 2001-2013 Devin Teske <dteske@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "string_m.h"

#include <stdio.h>

/* counts the number of occurrences of one string that appear in the source
 * string. Return value is the total count.
 *
 * An example use would be if you need to know how large a block of memory
 * needs to be for a replaceall() series.
 */
int
strcount(char *source, const char *find)
{
	char *p = source;
	u_int32_t n = 0;
	size_t flen;

	/* both parameters are required */
	if (!source || !find) return 0;

	/* cache length of find element */
	flen = strlen(find);
	if (!strlen(source) || !flen) return 0;

	/* loop until the end of the string */
	while (*p) {
		if (!strncmp(p, find, flen)) {
			p += flen; n++; /* found an instance */
		} else
			p++;
	}

	return n;
}

/* Replaces all occurrences of `find' in `source' with `replace'.
 *
 * You should not pass a string constant as the first parameter, it needs to be
 * a pointer to an allocated block of memory. The block of memory that source
 * points to should be large enough to hold the result. If the length of the
 * replacement string is greater than the length of the find string, the result
 * will be larger than the original source string. To allocate enough space for
 * the result, use the function strcount() declared above to determine the
 * number of occurrences and how much larger the block size needs to be.
 *
 * If source is not large enough, the application will crash. The return value
 * is the length (in bytes) of the result.
 *
 * When an error occurs, -1 is returned and the global variable errno is set
 * accordingly. Returns zero on success.
 */
int
replaceall(char *source, const char *find, const char *replace)
{
	char *p, *t, *temp;
	u_int32_t n = 0;
	size_t slen, flen, rlen;

	errno = 0; /* reset global error number */

	/* check that we have non-null parameters */
	if (!source) return 0;
	if (!find) return strlen(source);

	/* cache the length of the strings */
	slen = strlen(source);
	flen = strlen(find);
	rlen = replace ? strlen(replace) : 0;

	/* cases where no replacements need to be made */
	if (slen == 0 || flen == 0 || slen < flen) return slen;

	/* if replace is longer than find, we'll need to create a temp copy */
	if (rlen > flen) {
		temp = (char *)malloc(slen + 1);
		if (errno) return -1; /* could not allocate memory */
		strcpy(temp, source);
	} else
		temp = source;

	/* reconstruct the string with the replacements */
	p = source; t = temp; /* position elements */
 
	while (*t) {
		if (!strncmp(t, find, flen)) {
			/* found an occurrence */
			for (n = 0; replace && replace[n]; n++)
				*p++ = replace[n];
			t += flen;
		} else
			*p++ = *t++; /* copy character and increment */
	}

	/* terminate the string */
	*p = 0;

	/* free the temporary allocated memory */
	if (temp != source) free(temp);

	/* return the length of the completed string */
	return strlen(source);
}
